from web_engine.engine.abstract.AbstractParser import AbstractParser

from interface import implements
from lxml import html, etree
import requests
from urllib.parse import urljoin


class LxmlParser(AbstractParser):
    pass




from web_engine.interfaces.Query import Query
from web_engine.utils.JsonReader import JsonReader

from interface import implements
import os
import json


class JsonQuery(implements(Query)):

    def __init__(self, json_query_path: str = None, json_query: dict = None):
        if json_query_path is None:
            if json_query is None:
                raise Exception('Please supply a json_query_path or a json_query')
            self.__data = json_query
            self.__json_query_path = None
            self.__status = True
            self.__json_reader = JsonReader(self.__data)
        else:
            self.__data = None
            self.__json_query_path = None
            self.__status = False
            self.__json_reader = None
            self.__load(json_query_path)
            if self.__status:
                self.__json_reader = JsonReader(self.__data)

    def load(self, query_def: dict = None, json_query_path: str = None) -> Query:
        if query_def is None:
            return self.load(json_query_path)
        return JsonQuery(json_query=query_def)

    def __load(self, json_query_path: str) -> Query:
        if not os.path.exists(json_query_path):
            self.__status = False
            raise Exception("File {} not found".format(json_query_path))
        try:
            with open(json_query_path) as json_file:
                data = json.load(json_file)
                self.__json_query_path = json_query_path
                self.__data = data
                self.__status = True
                self.__json_reader = JsonReader(self.__data)
                return self
        except Exception as ex:
            self.__status = False
            raise Exception("Unable to read json file {}, {}".format(json_query_path, ex))

    @property
    def query_def(self):
        return self.__data

    @property
    def query_items(self):
        return {key: value for key, value in self.__data.items() if not key.startswith('_')}

    @property
    def status(self):
        return self.__status

    @property
    def prop(self) -> JsonReader:
        if not self.__status:
            return None
        return self.__json_reader

    @property
    def pre_filters(self) -> list:
        return [{key: value} for key, value in self.__data.items() if key.startswith('_pre_')]
        # result = list()
        # for key, value in self.__data.items():
        #     if key.startswith('_pre_'):
        #         result.append({key: value})
        # return result



from interface import Interface


class Engine(Interface):
    """
    Base interface for the engine that will do the main query processing and return the end result
    """

    def process(self) -> dict:
        pass




from interface import Interface
from web_engine.interfaces import Query


class Query(Interface):
    """
    Implement this interface to parse the query into executable steps
    """

    def load(self, query_def: dict = None, json_query_path: str = None) -> Query:
        pass

    @property
    def query_def(self) -> dict:
        pass

    @property
    def query_items(self):
        pass

    @property
    def status(self) -> bool:
        pass

    @property
    def prop(self):
        pass

    @property
    def pre_filters(self) -> list:
        pass



from web_engine.interfaces.Action import Action

import logging
from interface import implements


class FilterRemoveAction(implements(Action)):

    def __init__(self, receiver, log: logging = None):
        self._log = log
        self._receiver = receiver

    def execute(self, *args):
        if args is None or len(args) != 1:
            return
        xpath_selector: str = args[0]
        elements = self._receiver.driver.find_elements_by_xpath(xpath_selector)
        if elements is not None:
            for element in elements:
                try:
                    self._receiver.driver.execute_script("""
                    var element = arguments[0];
                    element.parentNode.removeChild(element);
                    """, element)
                except Exception as ex:
                    if self._log:
                        self._log.warn(ex)




from interface import Interface


class Scraper(Interface):

    def get(self):
        pass

    def get_doc(self, doc):
        pass

    def action_get(self, actions: list):
        pass

    def filter_get(self, filters: list):
        pass



from web_engine.interfaces.Engine import Engine
from web_engine.interfaces.Parser import Parser
from web_engine.interfaces.Query import Query

from interface import implements
import logging


class CoreEngine(implements(Engine)):

    def __init__(self, parser: Parser, query: Query, log: logging = None, context=None):
        if log is None:
            logging.basicConfig(level=logging.ERROR)
            self.log = logging.getLogger('Engine')
        else:
            self.log = log

        self.parser = parser
        self.query = query
        self.context = context

    def process(self) -> dict:
        if not self.query.status:
            return None

        doc = self.query.prop._doc.get()
        if doc:
            self.context = self.parser.get_doc_element(doc)

        if self.context is None:
            return None

        return self.__process()

    def __process(self) -> dict:
        result = dict()

        self.log.info("XPathEngine::__process -> query: {}".format(self.query.query_def))

        for key, value in self.query.query_items.items():
            self.log.info("XPathEngine::__process -> key, value {}, {}".format(key, value))
            type_value = type(value)
            self.log.info("XPathEngine::__process -> value type {}".format(type_value))

            if type_value == type(None):
                break
            if type_value in [str, int, float, bool]:
                result.update(self._process_item(key, value, self.context))
            if type_value == list:
                if result.get(key) is None:
                    result[key] = list()
                temp_return = self.__process_array(self.query.load(query_def=value[0]))
                result[key].extend(temp_return)
            if type_value == dict:
                x = CoreEngine(self.parser, self.query.load(query_def=value),
                               context=self.context, log=self.log)
                result.update({key: x.process()})

        self.log.info("XPathEngine::__process -> result {}".format(result))
        return result

    def __process_array(self, inner_query: Query) -> list:
        result = list()
        html_element = self.context

        if inner_query.prop._doc.get():
            return self._process_doc(inner_query)

        pre_filters = inner_query.pre_filters
        if len(pre_filters) > 0:
            html_element = self._apply_pre_filter(pre_filters)

        _base_path = inner_query.prop._base_path.get()
        if not _base_path:
            self.log.warning("XPathEngine::_process_array -> _base_path was not present")
            _base_path = "//*"
            # return result

        self.log.info("XPathEngine::_process_array -> xpath {}".format(_base_path))
        selector = self.parser.get_element(html_element, _base_path)
        if len(selector) == 0:
            self.log.warning("XPathEngine::_process_array -> selector length was 0")
            return result

        _key = inner_query.prop._key.get()

        for item in selector:
            self.log.info("XPathEngine::_process_array -> selector.item {}".format(item))

            inner_context = self.parser.get_inner_element(item)
            x = CoreEngine(self.parser, inner_query.load(inner_query.query_def), context=inner_context, log=self.log)
            if _key:
                result.append({_key: x.process()})
            else:
                result.append(x.process())

        if self.query.prop._follow.get():
            next_url = self.parser.get_element(html_element, self.query.prop._follow.get())

            if next_url is not None and len(next_url) > 0:

                if self.query.prop._follow_action.get():
                    action = self.query.prop._follow_action.get()
                    next_html_element = self.parser.get_doc_element_by_actions([action])
                    result.extend(CoreEngine(self.parser, inner_query, self.log, next_html_element).
                                  __process_array(inner_query))
                else:
                    next_url = self.parser.make_absolute_url(next_url[0])
                    next_html_element = self.parser.get_doc_element(next_url)
                    result.extend(CoreEngine(self.parser, inner_query, self.log, next_html_element).
                                  process_array(inner_query))

        self.log.info("XPathEngine::__process_array -> return result {}".format(result))
        return result

    def _process_item(self, key, value, html_element):

        self.log.info("XPathEngine::__process_item -> input key,value {}".format(key, value))

        self.log.info("XPathEngine::__process_item -> input html_elements {}".format(html_element))

        result = self.parser.get_items(value, html_element)

        self.log.info("XPathEngine::__process_item -> return result {}".format(result))
        return {key: result}

    def _process_doc(self, inner_query: Query) -> list():
        self.log.info("XPathEngine::_process_array -> doc {}".format(inner_query.prop.doc.get()))
        html_element = self.parser.get_doc_element(inner_query.prop._doc.get())
        del inner_query.query_def['_doc']
        inner_query = inner_query.load(query_def=inner_query.query_def)
        return CoreEngine(self.parser, inner_query, context=html_element, log=self.log).__process_array(inner_query)

    def _apply_pre_filter(self, filters: list):
        return self.parser.get_doc_element_by_filter(filters)



from web_engine.engine.JsonQuery import JsonQuery
import pytest
import pathlib


@pytest.fixture
def ex_config_reader():
    """Expected to raise a generic exception when file path is not found"""
    with pytest.raises(Exception):
        return JsonQuery(json_query_path=f"{pathlib.Path(__file__)}/NOFILE")


def test_reader(config_reader):
    """test that file has been loaded and element _doc is present"""
    doc = config_reader.prop._doc.get()
    print(doc)
    assert doc and len(doc) > 0


def test_no_file_reader(ex_config_reader):
    """test that a managed exception is thrown"""
    pass


def test_json_query_from_dict(dict_config_reader):
    """test that JsonQuery from dictionary has been loaded and element _doc is present"""
    assert dict_config_reader.prop._doc.get()


def test_query_items(config_reader: JsonQuery):
    """test that filters and actions are removed and only query_items are returned"""
    items = config_reader.query_items
    assert len(items) > 0
    assert "_key" not in items
    assert "_doc" not in items
    assert len([x for x in items if x.startswith("_")]) == 0


def test_pre_filter_items(dict_config_reader):
    """test  that _pre_ filters are returned"""
    items = JsonQuery(json_query=dict_config_reader.prop.reviews.get()[0]).pre_filters
    assert len(items) > 0, "Items was empty"
    assert "_pre_remove" in items[0].keys(), "Key _pre_remove not found in filters"




from web_engine.interfaces.Scraper import Scraper
from web_engine.interfaces.ActionRunner import ActionRunner
from web_engine.engine.actions.ClickAction import ClickAction
from web_engine.engine.actions.FilterRemoveAction import FilterRemoveAction
from web_engine.engine.config.ScraperConfig import ScraperConfig
import geckodriver_autoinstaller

from interface import implements
import time
import re
from urllib.parse import urljoin

from selenium import webdriver
from xvfbwrapper import Xvfb

geckodriver_autoinstaller.install()


class SeleniumScraper(implements(Scraper, ActionRunner)):

    def __init__(self, log, config: ScraperConfig):
        self.log = log
        self.config = config
        firefox_profile = webdriver.FirefoxProfile()
        firefox_profile.set_preference('permissions.default.image', self.config.img.get())
        firefox_profile.set_preference('dom.ipc.plugins.enabled.libflashplayer.so', 'false')

        if self.config.xvfb.get():
            self.display = Xvfb()
            self.display.start()

        self._driver = webdriver.Firefox(firefox_profile)

        self._driver.set_window_size(self.config.width.get(), self.config.height.get())
        self._driver.set_window_position(self.config.lat.get() or 0, self.config.lon.get() or 0)
        self.last_url = None
        self.last_page = None
        self._actions = {}
        self._history = []

        click = ClickAction(self, self.log)
        filter_remove = FilterRemoveAction(self, self.log)
        self.register('click', click)
        self.register('remove', filter_remove)

    @property
    def actions(self) -> dict:
        return self._actions

    def get_doc(self, doc):
        """
        get the html element from the url [doc]
        :param doc: the url to the page
        :return: html.from string content
        """
        self._driver.get(doc)
        page_source = self.driver.page_source
        self.last_url = doc
        self.last_page = page_source
        return self.last_page

    def get(self):
        page_source = self._driver.page_source
        self.last_page = page_source
        return self.last_page

    def action_get(self, actions: list):
        for x in actions:
            self.execute(x)
        return self.get()

    def filter_get(self, filters: list):
        for x in filters:
            self.execute_filter(x)
        return self.get()

    @property
    def history(self):
        return self._history

    def execute_filter(self, filter_def: dict):
        action_name, action_path = next(iter(filter_def.items()))
        action_name = action_name.strip().replace('_pre_', '')
        if action_name in self._actions.keys():
            self._history.append((time.time(), action_name))
            self._actions[action_name].execute(action_path)
        else:
            self.log.warn(f"Command [{action_name}] not recognised")

    def execute(self, action_composite: str):
        action_name, action_path = SeleniumScraper.__get_action(action_composite)
        action_name = action_name.strip()
        if action_name in self._actions.keys():
            self._history.append((time.time(), action_name))
            self._actions[action_name].execute(action_path)
        else:
            self.log.warn(f"Command [{action_name}] not recognised")

    @property
    def driver(self):
        return self._driver

    def make_absolute_url(self, url):
        return urljoin(self.last_url, url)

    @staticmethod
    def __get_action(action_composite):
        pattern = '{(.+?)}'
        matches = re.search(pattern, action_composite)
        if not matches:
            return None, None
        action_name = matches.group(1)
        action_xpath = re.sub(pattern, '', action_composite)
        return action_name, action_xpath

    def __del__(self):
        if self.driver is not None:
            try:
                time.sleep(10)  # wait for operations to complete before closing
                if hasattr(self, "_driver"):
                    self.driver.close()
                if self.config.xvfb.get():
                    self.display.stop()
            except Exception as ex:
                self.log.error(ex)
                pass




from web_engine.interfaces.Scraper import Scraper
from web_engine.interfaces.ActionRunner import ActionRunner

from interface import implements
import requests


class RequestScraper(implements(Scraper, ActionRunner)):

    def __init__(self, log):
        self.log = log

        self.last_url = None
        self.last_page = None
        self._actions = {}
        self._history = []

    @property
    def actions(self) -> dict:
        return self._actions

    def get_doc(self, doc):
        """
        get the html element from the url [doc]
        :param doc: the url to the page
        :return: html.from string content
        """
        with requests.get(doc) as web_request:
            page_source = web_request.content
            self.last_url = doc
            self.last_page = page_source
            return self.last_page

    def get(self):
        return self.last_page

    def action_get(self, actions: list):
        self.log.error(f"[{type(self)}] Does not support actions. Please use another Scraper that supports actions")
        self.execute(x for x in actions)
        return self.get()

    def filter_get(self, filters: list):
        self.log.error(f"[{type(self)}] Does not support filters. Please use another Scraper that supports filters")
        self.execute(x for x in filters)
        return self.get()

    def execute(self, action_composite: str):
        action_name, action_path = RequestScraper.__get_action(action_composite)
        self.log.warn(f"Command [{action_name}] not recognised")

    @property
    def driver(self):
        return None

    @property
    def history(self):
        return self._history





from interface import Interface
from interface import interface

import logging


class Action(Interface):
    """
    Implement this interface to support action on the scraping engine
    """

    def execute(self, *args):
        pass







#!/usr/local/bin/python3

from web_engine.engine.JsonQuery import JsonQuery
from web_engine.engine.CoreEngine import CoreEngine
from web_engine.engine.LxmlParser import LxmlParser
from web_engine.engine.SeleniumScraper import SeleniumScraper
from web_engine.engine.RequestScraper import RequestScraper
from web_engine.engine.config.ScraperConfig import ScraperConfig

from pprint import pprint as pp

import argparse
import logging
import json
import sys

if not sys.warnoptions:
    import warnings
    warnings.simplefilter("ignore")

parser = argparse.ArgumentParser(description='Web Scrap Engine for semi-structured web data retrieval using JSON '
                                             'query constructs')

# Required positional argument
parser.add_argument('-q', '--query', type=str,
                    help='JSON query', action="store", dest="query")

parser.add_argument('-e', '--engine', type=str, nargs="?",
                    help='Engine: use [lxml] for parser engine (default), [selenium] for action based web scraping',
                    action="store", dest="engine", default="selenium")

parser.add_argument('-ht', '--height', type=int, nargs="?",
                    help="specify the browser window height (default is 800, only used with Selenium engine)",
                    action="store", dest="height", default=800
                    )

parser.add_argument('-wh', '--width', type=int, nargs="?",
                    help="specify the browser window width (default is 1280, only used with Selenium engine)",
                    action="store", dest="width", default=1280
                    )

parser.add_argument('-lat', '--lat', type=int, nargs="?",
                    help="Latitude (not specified by default)",
                    action="store", dest="lat")

parser.add_argument('-lon', '--lon', type=int, nargs="?",
                    help="Longitude (not specified by default)",
                    action="store", dest="lon")

parser.add_argument('-img', '--img', help="Load images",
                    action="store_true", dest="img")

parser.add_argument('-l', '--log',
                    help='Set flag to True to see verbose logging output',
                    action="store_true", dest="log")

parser.add_argument('-xvfb', '--xvfb',
                    help='Set flag to False to see Firefox when using Selenium engine',
                    action="store_true", dest="xvfb")

sys.setrecursionlimit(1000)


def main(argv):
    args = parser.parse_args()
    log_level = logging.ERROR
    if args.log:
        log_level = logging.NOTSET

    logging.basicConfig(filename='dr-web-engine.log', level=logging.ERROR)
    root_log = logging.getLogger()
    root_log.handlers = []
    root_log.addHandler(logging.FileHandler('dr-web-engine.log.1', mode='a', encoding=None, delay=False))

    log = logging.getLogger(__name__)
    log.setLevel(log_level)
    log.handlers = []
    log.addHandler(logging.FileHandler('dr-web-engine.log', mode='a', encoding=None, delay=False))

    if args.query is None:
        parser.print_help()
        return

    query = JsonQuery(json_query_path=args.query)

    scraper = RequestScraper(log)
    if args.engine == 'selenium':
        scraper_config = ScraperConfig({"xvfb": args.xvfb,
                                        "lon": args.lon or None,
                                        "lat": args.lat or None,
                                        "width": args.width,
                                        "height": args.height,
                                        "img": 1 if args.img else 2
                                        })
        scraper = SeleniumScraper(log, scraper_config)

    html_parser = LxmlParser(scraper, log)

    engine = CoreEngine(html_parser, query, log)

    result = engine.process()

    # sys.stdout.write(json.dumps(result))
    # sys.stdout.flush()

    print(json.dumps(result))

    # pp(result)

    sys.exit(0)

    # print(json.dumps(result))


if __name__ == "__main__":

    # sys.argv.append("-q")
    # sys.argv.append("../test/trader.extract.json")

    # sys.argv.append("--log")

    # sys.argv.append("-wh")
    # sys.argv.append("640")
    #
    # sys.argv.append("-ht")
    # sys.argv.append("640")

    # sys.argv.append("-lon")
    # sys.argv.append("20")
    #
    # sys.argv.append("-lat")
    # sys.argv.append("20")
    #
    # sys.argv.append("-img")
    # sys.argv.append("True")

    # sys.argv.append("-e")
    # sys.argv.append("selenium")

    main(sys.argv)




from web_engine.interfaces.Action import Action

from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium import webdriver

from interface import implements
import logging


class ClickAction(implements(Action)):

    def __init__(self, receiver, log: logging = None):
        self._log = log
        self._receiver = receiver

    def execute(self, *args):
        if args is None or len(args) != 1:
            return
        xpath_selector: str = args[0]
        wait = WebDriverWait(self._receiver.driver, 10)
        elem = wait.until(EC.element_to_be_clickable((By.XPATH, xpath_selector)))
        elem.click()




from interface import Interface
from interface import interface

from web_engine.interfaces.Action import Action
import re


class ActionRunner(Interface):
    """
    Implement this interface on objects that can execute Actions
    """

    @property
    def actions(self) -> dict:
        pass

    @property
    def history(self):
        pass

    @interface.default
    def register(self, action_name: str, action: Action):
        self.actions[action_name] = action

    def execute(self, action_composite: str):
        pass



from web_engine.engine.CoreEngine import CoreEngine

import pytest
from unittest.mock import Mock

parser = Mock()
parser.get_doc_element.return_value = "<html><body>Hello</body></html>"
parser.get_doc_element_by_filter.return_value = "<html><body>Hello</body></html>"
parser.get_element.return_value = []


@pytest.fixture
def engine(dict_config_reader):
    core_engine = CoreEngine(parser=parser, query=dict_config_reader, log=None, context=None)
    return core_engine


def test_main_process(engine):
    result = engine.process()
    assert engine.parser.get_doc_element.called, "Method get_doc_element was not called"
    assert engine.parser.get_doc_element_by_filter.called, "Method get_doc_element_by_filter not called"
    assert engine.parser.get_element.called, "Method get_element not called"
    assert result


def test_process_doc(engine, simple_config):
    main_content = "<html><body><h1>Expected Result</h1></body><html>"
    engine.parser.get_doc_element.return_value = main_content
    engine.parser.get_element.return_value = [main_content]
    engine.parser.get_items.return_value = "Expected Result"
    result = engine._process_doc(simple_config)
    assert len(result) == 1, f"Wrong array length: expected 1, return {len(result)}"
    assert result[0]["h1"] == "Expected Result", "Failed to get the expected h1 return"


def test_process_item(engine, simple_config):
    engine.parser.get_items.side_effect = lambda x, _: "Expected Result" if x == simple_config.prop.h1.get() else ""
    result = engine._process_item("h1", simple_config.prop.h1.get(), "")
    assert result["h1"] == "Expected Result"






class JsonReader:

    def __init__(self, json_object, cache=None):
        self.json_object = json_object
        self._cache = cache or []

    # Final method call
    def get(self):
        temp_value = self.json_object
        for item in self._cache:
            if not isinstance(item, int) and item not in temp_value:
                return None
            temp_value = temp_value[item]
        return temp_value

    # Build the cache, and handle special cases
    def _(self, name):
        # Enables method chaining
        return JsonReader(self.json_object, self._cache + [name])

    # Reflection
    def __getattr__(self, name):
        return self._(name)

    # Called with the object is deleted
    def __del__(self):
        self.json_object = None
        self._cache = None

    def __getitem__(self, item):
        return self._(item)




from interface import Interface


class Parser(Interface):
    """
    Implement this interface to support multiple parsing methods for example XPath and DOM and/or to use different
    parsing engines
    """

    def get_doc_element(self, doc):
        pass

    def get_element(self, html_element, xpath):
        pass

    def get_items(self, value: str, html_element) -> list:
        pass

    def get_inner_element(self, element, actions=None):
        pass

    def make_absolute_url(self, url):
        pass

    def get_doc_element_by_actions(self, actions: list):
        pass

    def get_doc_element_by_filter(self, filters: list):
        pass



from web_engine.interfaces import Action


class ActionBuilder:

    @staticmethod
    def build(action_string: str) -> Action:
        pass

    @staticmethod
    def build(action_strings: dict) -> Action:
        return [ActionBuilder.build(x) for _, x in action_strings.items()]



from web_engine.engine.JsonQuery import JsonQuery
from web_engine.engine.CoreEngine import CoreEngine
from web_engine.interfaces.Parser import Parser

from unittest.mock import Mock

import pytest
import pathlib


@pytest.fixture
def config_reader():
    """JsonQuery loaded from  trader.extract.json file in the test directory path"""
    return JsonQuery(json_query_path=f"{pathlib.Path(__file__).parent.absolute()}/trader.extract.json")


@pytest.fixture
def simple_config():
    json_query = {
        "_doc": "https://www.checkatrade.com/trades/WayreHouseElectricalServices",
        "h1": "//h1/text()"
    }
    return JsonQuery(json_query=json_query)


@pytest.fixture
def dict_config_reader():
    """JsonQuery loaded from dictionary"""
    json_query = {
        "_doc": "https://www.checkatrade.com/trades/WayreHouseElectricalServices",
        "data": {
            "ld_data": "//head/script[@type=\"application/ld+json\"][1]"
        }, "reviews": [{
            "_doc": "https://www.checkatrade.com/trades/WayreHouseElectricalServices/reviews",
            "_base_path": "//div[contains(@class, 'ReviewsPage__Content')]//div[contains(@class, 'ReviewsItem__Wrapper')]",
            "_key": "review",
            "_pre_remove": "//*[contains(@class,'alert-box')]",
            "_follow": "//a[contains(@class,\"Chevrons__Wrapper\")][2]/@href",
            "_follow_action": "//a[contains(@class,\"Chevrons__Wrapper\")][2]{click }",
            "title": "//h3[contains(@class, 'ReviewsItem__Title')]",
            "score": "//*[name()='svg']//text()[normalize-space()]",
            "verified": "//div[contains(@class, 'ReviewsItem__Verified')]/text()[normalize-space()]",
            "content": "//p[contains(@class, 'ReviewsItem__P')]",
            "review_by": "//div[contains(@class, 'ReviewsItem__Byline')]/text()[normalize-space()]"
        }]}

    return JsonQuery(json_query=json_query)






import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
        name="dr-web-engine",
        version="0.3.2.1Beta",
        author="Ylli Prifti",
        author_email="ylli@dcs.bbk.ac.uk",
        description="Data Retrieval Web Engine - Queryable Web Scrap engine build on python based on lxml and "
                    "Selenium package and using JSON as query construct.",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/ylliprifti/dr-web-engine",
        packages=setuptools.find_packages(),
        install_requires=['requests', 'argparse', 'xvfbwrapper', 'selenium', 'python_interface',
                          'lxml', 'geckodriver_autoinstaller'],
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        python_requires='>=3.6',
)



from web_engine.interfaces.Parser import Parser
from web_engine.interfaces.Scraper import Scraper

from interface import implements
from lxml import html, etree
import requests
from urllib.parse import urljoin

import logging


class AbstractParser(implements(Parser)):

    def __init__(self, scraper: Scraper, log: logging = None):
        self.log = log
        self.last_tree: etree = None
        self.last_element: html.HtmlElement = None
        self.last_inner_element = None
        self.last_items: list = None
        self.last_url: str = None
        self.scraper = scraper

    def get_doc_element(self, doc):
        '''
        get the html element from the url [doc]
        :param doc: the url to the page
        :return: html.from string content
        '''
        return self.scraper.get_doc(doc)

    def get_items(self, value: str, html_element) -> list:
        if type(html_element) is str:
            html_element = html.fromstring(html_element)
        if type(html_element) is bytes:
            html_element = html.fromstring(html_element.decode("utf-8"))
        xpath_result = html_element.xpath(value)
        result = list()
        for x in xpath_result:
            if type(x) == html.HtmlElement:
                result.append(x.text)
            else:
                result.append(x)

        self.last_items = result
        return self.last_items

    def get_element(self, html_element, xpath):
        if type(html_element) is str:
            html_element = html.fromstring(html_element)
        if type(html_element) is bytes:
            html_element = html.fromstring(html_element.decode("latin1"))
        self.last_element = html_element.xpath(xpath)
        return self.last_element

    def get_inner_element(self, element, actions=None):
        self.last_inner_element = html.fromstring(etree.tostring(element, pretty_print=True))
        return self.last_inner_element

    def make_absolute_url(self, url):
        return urljoin(self.last_url, url)

    def get_doc_element_by_actions(self, actions: list):
        return self.scraper.action_get(actions)

    def get_doc_element_by_filter(self, filters: list):
        return self.scraper.filter_get(filters)



